@@ -1,11 +1,11 @@
 use strict;
 use warnings;
-package Mojolicious::Plugin::OpenAPI::Modern; # git description: 9d91507
+package Mojolicious::Plugin::OpenAPI::Modern; # git description: v0.001-10-gc44300b
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Mojolicious plugin providing access to an OpenAPI document and parser
 # KEYWORDS: validation evaluation JSON Schema OpenAPI Swagger HTTP request response
 
-our $VERSION = '0.001';
+our $VERSION = '0.002';
 
 use 5.020;
 use if "$]" >= 5.022, experimental => 're_strict';
@@ -16,11 +16,16 @@ use Mojo::Base 'Mojolicious::Plugin', -signatures;
 use Feature::Compat::Try;
 use YAML::PP;
 use Path::Tiny;
-use JSON::MaybeXS;
+use Mojo::JSON 'decode_json';
 use Safe::Isa;
 use OpenAPI::Modern 0.022;
 use namespace::clean;
 
+# we store data in two places: on the app (persistent storage, for the OpenAPI::Modern object
+# itself) and in the controller stash: per-request data like the path info and extracted path items.
+
+# the first is $app->openapi or $c->openapi
+# the second is $c->stash('openapi') which will be initialized to {} on first use.
 sub register ($self, $app, $config) {
   my $stash = Mojo::Util::_stash(openapi => $app);
 
@@ -34,8 +39,7 @@ sub register ($self, $app, $config) {
         $schema = YAML::PP->new(boolean => 'JSON::PP')->load_file($config->{document_filename}),
       }
       elsif ($config->{document_filename} =~ /\.json$/) {
-        $schema = JSON::MaybeXS->new(allow_nonref => 1, utf8 => 1)->decode(
-          path($config->{document_filename})->slurp_raw);
+        $schema = decode_json(path($config->{document_filename})->slurp_raw);
       }
       else {
         die 'Unsupported file format in filename: ', $config->{document_filename};
@@ -46,21 +50,32 @@ sub register ($self, $app, $config) {
     }
 
     my $openapi = OpenAPI::Modern->new(
-        openapi_uri    => $config->{document_filename} // '',
-        openapi_schema => $schema,
+      openapi_uri    => $config->{document_filename} // '',
+      openapi_schema => $schema,
     );
 
     # leave room for other keys in our localized stash
     $stash->{openapi} = $openapi;
   }
   catch ($e) {
-    die 'Cannot load OpenAPI document: ', $e if not $e->$_isa('JSON::Schema::Modern::Result');
-    my $encoder = JSON::MaybeXS->new(canonical => 1, pretty => 1, utf8 => 0);
-    $encoder->indent_length(2) if $encoder->can('indent_length');
-    die $encoder->encode($e->TO_JSON);
+    die 'Cannot load OpenAPI document: ', $e;
   }
 
   $app->helper(openapi => sub ($c) { $stash->{openapi} });
+
+  $app->helper(validate_request => \&_validate_request);
+  $app->helper(validate_response => \&_validate_response);
+}
+
+sub _validate_request ($c) {
+  my $options = $c->stash->{openapi} //= {};
+  return $c->openapi->validate_request($c->req, $options);
+}
+
+sub _validate_response ($c) {
+  my $options = $c->stash->{openapi} //= {};
+  local $options->{request} = $c->req;
+  return $c->openapi->validate_response($c->res, $options);
 }
 
 1;
@@ -77,7 +92,7 @@ Mojolicious::Plugin::OpenAPI::Modern - Mojolicious plugin providing access to an
 
 =head1 VERSION
 
-version 0.001
+version 0.002
 
 =head1 SYNOPSIS
 
@@ -121,12 +136,66 @@ Instantiates an L<OpenAPI::Modern> object and provides an accessor to it.
 These methods are made available on the C<$c> object (the invocant of all controller methods,
 and therefore other helpers).
 
-=for stopwords openapi
+=for stopwords openapi operationId
 
 =head2 openapi
 
 The L<OpenAPI::Modern> object.
 
+=head2 validate_request
+
+  my $result = $c->openapi->validate_request;
+
+Passes C<< $c->req >> to L<OpenAPI::Modern/validate_request> and returns the
+L<JSON::Schema::Modern::Result>.
+
+Note that the matching L<Mojo::Routes::Route> object for this request is I<not> used to find the
+OpenAPI path-item that corresponds to this request: only information in the request URI itself is
+used (although some information in the route may be used in a future feature).
+
+=head2 validate_response
+
+  my $result = $c->openapi->validate_response;
+
+Passes C<< $c->res >> and C<< $c->req >> to L<OpenAPI::Modern/validate_response> and returns the
+L<JSON::Schema::Modern::Result>.
+
+Can only be called in the areas of the dispatch flow where the response has already been rendered; a
+good place to call this would be in an L<after_dispatch|Mojolicious/after_dispatch> hook.
+
+Note that the matching L<Mojo::Routes::Route> object for this request is I<not> used to find the
+OpenAPI path-item that corresponds to this request and response: only information in the request URI
+itself is used (although some information in the route may be used in a future feature).
+
+=head1 STASH VALUES
+
+This plugin stores all its data under the C<openapi> hashref, e.g.:
+
+  my $operation_id = $c->stash->{openapi}{operation_id};
+
+Keys starting with underscore are for I<internal use only> and should not be relied upon to behave
+consistently across release versions. Values that may be used by controllers and templates are:
+
+=over 4
+
+=item *
+
+C<path_template>: Set by the first call to L</validate_request> or L</validate_response>. A string representing the request URI, with placeholders in braces (e.g. C</pets/{petId}>); see L<https://spec.openapis.org/oas/v3.1.0#paths-object>.
+
+=item *
+
+C<path_captures>: Set by the first call to L</validate_request> or L</validate_response>. A hashref mapping placeholders in the path to their actual values in the request URI.
+
+=item *
+
+C<operation_id>: Set by the first call to L</validate_request> or L</validate_response>. Contains the corresponding L<operationId|https://swagger.io/docs/specification/paths-and-operations/#operationid> of the current endpoint.
+
+=item *
+
+C<method>: Set by the first call to L</validate_request> or L</validate_response>. The HTTP method used by the request, lower-cased.
+
+=back
+
 =head1 SEE ALSO
 
 =over 4
